using System;
using System.Text;
using System.Linq;
using System.Collections.Generic;
using System.IO;
using System.Diagnostics;

/// <summary>
/// Guru 5-6. Gives the costliest route of a upper triangle matrix from top to bottom.
/// This solution utilizes ``dynamic programming''.
/// </summary>
public class PisinReitti
{
    /// <summary>
    /// Pääohjelma.
    /// </summary>
    public static void Main()
    {
        // Määritellään tiedostopolku vakioksi
        const string FILEPATH = @"tree.txt";
        const string FILEPATH_SIMPLE = @"tree_simple.txt";

        // Avataan tiedosto ja kirjoitetaan sen sisältö ruudulle
        string[] luetutRivit = File.ReadAllLines(FILEPATH);
        int[,] values = new int[luetutRivit.Length - 1, luetutRivit.Length - 1];

        for (int i = 1; i < luetutRivit.Length; i++)
        {
            string[] rivi = luetutRivit[i].Split(' ');
            for (int j = 0; j < rivi.Length; j++)
            {
                values[i - 1, j] = int.Parse(rivi[j]);
            }

        }

        Stopwatch sw = Stopwatch.StartNew();
        Console.WriteLine(MaxCost(values));
        sw.Stop();
        double ms = sw.Elapsed.TotalMilliseconds;
        Console.WriteLine(ms + " ms");

        /*
        // Naive way
        int solutions = 0;
        sw = Stopwatch.StartNew();
        //Do Stuff
        Console.WriteLine(MaxCost(values, 0, 0, ref solutions));
        Console.WriteLine(solutions);
        sw.Stop();
        ms = sw.Elapsed.TotalMilliseconds;
        Console.WriteLine(ms + " ms");
        */

    }

    /// <summary>
    /// Gives the max cost of (i, j)
    /// </summary>
    /// <param name="values">Values</param>
    /// <param name="costs">Costs</param>
    /// <param name="row">row</param>
    /// <param name="col">col</param>
    /// <returns>Cost</returns>
    public static int Cost(int[,] values, int[,] costs, int row, int col)
    {
        if (row == 0) return values[row, col];
        if (col == 0) return values[row, col] + costs[row - 1, col];
        // if (col > row) return 0;
        if (col == row) return values[row, col] + costs[row - 1, col - 1];
        return values[row, col] + Math.Max(costs[row - 1, col], costs[row - 1, col - 1]);
    }


    /// <summary>
    /// Gives the sum of the costliest route.
    /// </summary>
    /// <param name="values">Value matrix</param>
    /// <returns>Costliest route</returns>
    /// <example>
    /// <pre name="test">
    /// int[,] values;
    /// values = new int[,] {         {89,  0,  0,  0,  0}, 
    ///                             {70,  2,  0,  0,  0}, 
    ///                           {39, 49, 87,  0,  0}, 
    ///                         {43, 93, 68, 64,  0}, 
    ///                       {87, 75, 69, 18, 44}};
    /// OutrunPaths.MaxCost(values) === 376;
    /// </pre>
    /// </example>
    public static int MaxCost(int[,] values)
    {
        int[,] costs = Costs(values);
        int[] costliest = RouteFromIndices(values, IndicesOfCostliestRoute(values, costs));
        return Sum(costliest);
    }

    /// <summary>
    /// Gives the sum of an array.
    /// </summary>
    /// <param name="values">Int values</param>
    /// <returns>Sum</returns>
    private static int Sum(int[] values)
    {
        int sum = 0;
        for (int i = 0; i < values.Length; i++)
        {
            sum += values[i];
        }
        return sum;
    }



    /// <summary>
    /// Returns costs for routes. This is the naïve way 
    /// that takes a lot of time.
    /// </summary>
    /// <param name="values"></param>
    /// <returns></returns>
    /// <example>
    /// <pre name="test">
    /// int[,] values ={ {  5,  0,  0,  0 } 
    ///                 ,{ 10, 15,  0,  0 }
    ///                 ,{ 10, 10, 15,  0 }
    ///                 ,{ 30, 10, 13, 10} } ;
    /// int solutions = 0;                 
    /// MaxCost(values, 0, 0, ref solutions) === 55;
    /// values = new int[,] {         {89,  0,  0,  0,  0}, 
    ///                             {70,  2,  0,  0,  0}, 
    ///                           {39, 49, 87,  0,  0}, 
    ///                         {43, 93, 68, 64,  0}, 
    ///                       {87, 75, 69, 18, 44}};
    /// MaxCost(values, 0, 0, ref solutions) === 376;
    /// </pre>
    /// </example>
    public static int MaxCost(int[,] values, int i, int j, ref int solutions)
    {
        solutions++;
        //Console.WriteLine(solutions);
        if (i == values.GetLength(0) - 1) return values[i, j];
        int cost = values[i, j] + Math.Max(MaxCost(values, i + 1, j, ref solutions), MaxCost(values, i + 1, Math.Min(i, j + 1), ref solutions));
        return cost;
    }

    /// <summary>
    /// Gives costs for routes in a int matrix.
    /// </summary>
    /// <param name="values"></param>
    /// <returns></returns>
    /// <example>
    /// <pre name="test">
    /// int[,] values ={ {  5,  0,  0,  0 } 
    ///                 ,{ 10, 15,  0,  0 }
    ///                 ,{ 10, 10, 15,  0 }
    ///                 ,{ 30, 10, 13, 10} } ;
    /// int[,] costs = OutrunPaths.Costs(values);
    /// Helpers.Jonoksi(costs," ","{0}",", ") === "5 0 0 0, 15 20 0 0, 25 30 35 0, 55 40 48 45";
    /// values = new int[,] { {89,  0,  0,  0,  0}, 
    ///                     {70,  2,  0,  0,  0}, 
    ///                   {39, 49, 87,  0,  0}, 
    ///                 {43, 93, 68, 64,  0}, 
    ///               {87, 75, 69, 18, 44}};
    /// costs = OutrunPaths.Costs(values);
    /// Helpers.Jonoksi(costs," ","{0}",", ") === "89 0 0 0 0, 159 91 0 0 0, 198 208 178 0 0, 241 301 276 242 0, 328 376 370 294 286";
    /// </pre>
    /// </example>
    public static int[,] Costs(int[,] values)
    {
        int[,] costs = new int[values.GetLength(0), values.GetLength(1)];
        for (int i = 0; i < values.GetLength(0); i++)
        {
            for (int j = 0; j <= i; j++)
            {
                costs[i, j] = Cost(values, costs, i, j);
            }
        }
        return costs;
    }


    /// <summary>
    /// Route (indices) to costliests in the values matrix.
    /// </summary>
    /// <param name="values"></param>
    /// <param name="costs"></param>
    /// <returns></returns>
    /// <example>
    /// <pre name="test">
    /// int[,] values ={ {        5,  0,  0,  0 } 
    ///                 ,{     10, 15,  0,  0 }
    ///                 ,{   10, 10, 15,  0 }
    ///                 ,{ 30, 10, 13, 10} } ;
    /// int[,] costs = {        { 5,   0,  0,  0 }
    ///                 ,    { 15, 20,  0,  0 }
    ///                 ,  { 25, 30, 35,  0 } 
    ///                 ,{ 55, 40, 48, 45} };
    /// int[] costliestIndices = OutrunPaths.IndicesOfCostliestRoute(values, costs);
    /// int[] costliest = OutrunPaths.RouteFromIndices(values, costliestIndices);
    /// String.Join(" ", costliest) === "5 10 10 30"
    /// values = new int[,] {         {89,  0,  0,  0,  0}, 
    ///                             {70,  2,  0,  0,  0}, 
    ///                           {39, 49, 87,  0,  0}, 
    ///                         {43, 93, 68, 64,  0}, 
    ///                       {87, 75, 69, 18, 44}};
    /// costs  = new int[,] {         {89,   0,   0,   0,   0}, 
    ///                             {159, 91,   0,   0,   0}, 
    ///                           {198, 208, 178,  0,  0}, 
    ///                         {241, 301, 276, 242,  0}, 
    ///                       {328, 376, 370, 294, 286}};
    /// costliestIndices = IndicesOfCostliestRoute(values, costs);
    /// String.Join(" ", costliestIndices) === "0 0 1 1 1";
    /// costliest = RouteFromIndices(values, costliestIndices);
    /// String.Join(" ", costliest) === "89 70 49 93 75";
    /// </pre>
    /// </example>
    public static int[] IndicesOfCostliestRoute(int[,] values, int[,] costs)
    {
        int[] indices = new int[values.GetLength(0)];
        int row = costs.GetLength(0) - 1; // starting from the last row
        int startingCol = 0; // column that has the biggest value in a row
        int biggest = int.MinValue; // biggest value found
        for (int col = 0; col < costs.GetLength(1); col++) // this is only for the last row
        {
            if (costs[row, col] > biggest)
            {
                biggest = costs[row, col];
                startingCol = col;
            }
        }

        indices[row] = startingCol;

        while (row > 0)
        {
            if (startingCol == 0) { indices[row] = startingCol; }
            else if (startingCol == row) { indices[row] = startingCol; startingCol--; }

            // checking the row above
            else if (costs[row - 1, startingCol - 1] > costs[row - 1, startingCol]) indices[row] = startingCol--;
            else indices[row] = startingCol;
            row--;
        }

        indices[0] = 0;

        return indices;
    }

    /// <summary>
    /// Gives route from indices
    /// </summary>
    /// <param name="values"></param>
    /// <param name="indices"></param>
    /// <returns></returns>
    /// <example>
    /// <pre name="test">
    /// </pre>
    /// </example>
    public static int[] RouteFromIndices(int[,] values, int[] indices)
    {
        int[] route = new int[values.GetLength(0)];
        for (int i = 0; i < indices.GetLength(0); i++)
        {
            route[i] = values[i, indices[i]];
        }
        return route;
    }

}
